local auth   = require("ssso_auth")
local conf   = require("ssso_config")
local crypto = require("ssso_crypto")
local log    = require("ssso_log")
local nginx  = require("ssso_nginx")
local sites  = require("ssso_sites")
local util   = require("ssso_util")

local root = ""

local function set_root(prefix)
  root = prefix .. "/login/"
end

local function contents(relative)
  local file = assert(io.open(root .. relative, "r"), "Cannot open login file " .. root .. relative)
  local data = file:read("*all")
  file:close()
  return data
end

local function inject_data(html, req_data, warnings)
  local paras
  local cause = ({
    ["401"] = "Credentials are needed to access this resource.",
    ["403"] = "Different credentials might grant access to this resource.",
  })[req_data.query_params.cause or ""]
  if cause then
    paras = '<p class="info">' .. util.str_to_html(cause) .. "</p>"
  else
    paras = ""
  end
  if warnings then
    for _, m in ipairs(warnings) do
      paras = paras .. '<p class="warning">' .. util.str_to_html(m) .. "</p>"
    end
  end
  if paras ~= ""  then
    html = html:gsub("<!%-%-MESSAGES%-%->", '<section id="messages">' .. paras .. "</section>", 1)
  end
  local back_url = req_data.query_params.back
  if not back_url then
    back_url = ""
  end
  html = html:gsub("BACK_URL", util.str_to_html(back_url))
  return html
end

local function check_credentials_and_get_profile(user, password)
  local user_data = auth.read_user(user, password)
  if user_data then
    log.debug("Credentials accepted for user " .. user_data.name)
    return sites.class__profile:build_from_conf(user, password, user_data.name, user_data.email)
  else
    return nil
  end
end

local function check_login(req_data)
  req_data = req_data:with_post_parameters()
  local user = req_data.query_params["login"] or ""
  local password = req_data.query_params["password"] or ""
  local profile = check_credentials_and_get_profile(user, password)
  if profile then
    log.debug("Building JWS")
    local jws, tslimit = crypto.get_jws_and_tslimit(profile)
    if not jws then
      return nginx.answer_unexpected_error()
    end
    nginx.set_jws_cookie(jws, tslimit)
    local back_url = req_data.query_params.back
    log.debug("Redirecting")
    if back_url then
      return nginx.redirect_to_page(back_url)
    else
      return nginx.redirect_to_portal()
    end
  else
    log.debug("Credentials rejected")
    local html = inject_data(contents("login.html"), req_data, {
      "These credentials were rejected."
    })
    return nginx.return_contents(html, "text/html; charset=UTF-8")
  end
end

local function answer_request(req_data)
  local login = conf.get_sso_prefix() .. "/login"
  if req_data:is(login) then
    if req_data:has_method("POST") then
      log.info("Checking login")
      return check_login(req_data)
    else
      local html = inject_data(contents("login.html"), req_data, nil)
      return nginx.return_contents(html, "text/html; charset=UTF-8")
    end
  elseif req_data:is(login .. ".css") then
    return nginx.return_contents(contents("login.css"), "text/css; charset=UTF-8")
  elseif req_data:is(login .. ".js") then
    return nginx.return_contents(contents("login.js"), "application/javascript; charset=UTF-8")
  else
    log.info("Unknown login file")
    return nginx.answer_not_found(req_data)
  end
end

return {
  answer_request = answer_request,
  check_credentials_and_get_profile = check_credentials_and_get_profile, -- TODO: test
  set_root = set_root,
}
