local lu = require("luaunit")
local ngx = require("ngx")
local conf = require("ssso_config")
local login = require("ssso_login")
local ng = require("ssso_nginx")

local here = debug.getinfo(1).source:sub(2, -17)
conf.load_conf(here)
login.set_root(here)

function test_get_login_url_returns_html_with_back_url_substitution()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "GET"
  ngx.var.request_uri = "/ssso/login?back=/somewhere"
  local r = ng.class__request:current()
  local expected = [[<html><head>
  <link href="login.css">
  <script src="login.js"></script>
</head><body>
  <!--MESSAGES-->
  <input value="/somewhere">
</body></html>
]]
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "text/html; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_login_css_url_returns_css()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "BLABLA"
  ngx.var.request_uri = "/ssso/login.css"
  local r = ng.class__request:current()
  local expected = "/*CSS*/\n"
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "text/css; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_login_js_url_returns_js()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "BLABLA"
  ngx.var.request_uri = "/ssso/login.js"
  local r = ng.class__request:current()
  local expected = "//JS\n"
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "application/javascript; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_unknown_login_url_returns_404()
  -- given
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "BLABLA"
  ngx.var.request_uri = "/ssso/login/unknown"
  local r = ng.class__request:current()
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 404)
  lu.assertNil(ngx.say())
end

function test_get_login_url_with_cause_401_returns_html_with_associated_message()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "GET"
  ngx.var.request_uri = "/ssso/login?cause=401"
  local r = ng.class__request:current()
  local expected = [[<html><head>
  <link href="login.css">
  <script src="login.js"></script>
</head><body>
  <section id="messages"><p class="info">Credentials are needed to access this resource.</p></section>
  <input value="">
</body></html>
]]
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "text/html; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_get_login_url_with_cause_403_returns_html_with_associated_message()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.var.request_method = "GET"
  ngx.var.request_uri = "/ssso/login?cause=403"
  local r = ng.class__request:current()
  local expected = [[<html><head>
  <link href="login.css">
  <script src="login.js"></script>
</head><body>
  <section id="messages"><p class="info">Different credentials might grant access to this resource.</p></section>
  <input value="">
</body></html>
]]
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "text/html; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_post_login_url_with_wrong_credentials_returns_html_with_associated_message()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.reset_post_var()
  ngx.var.request_method = "POST"
  ngx.var.request_uri = "/ssso/login"
  local r = ng.class__request:current()
  local expected = [[<html><head>
  <link href="login.css">
  <script src="login.js"></script>
</head><body>
  <section id="messages"><p class="warning">These credentials were rejected.</p></section>
  <input value="">
</body></html>
]]
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, 200)
  lu.assertEquals(ngx.say(), expected)
  lu.assertEquals(ngx.header["Content-Type"], "text/html; charset=UTF-8")
  lu.assertEquals(ngx.header["Content-Length"], tostring(#expected))
end

function test_post_login_url_with_good_credentials_redirects_to_portal_with_session_cookie()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.reset_post_var()
  ngx.var.request_method = "POST"
  ngx.var.request_uri = "/ssso/login"
  ngx.post_var.login = "goodlogin"
  ngx.post_var.password = "goodpassword"
  local r = ng.class__request:current()
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, "307:https://my-domain.tld/ssso/portal")
  lu.assertNil(ngx.say())
  lu.assertNil(ngx.header["Set-Cookie"].link)
  lu.assertStrMatches(ngx.header["Set-Cookie"].v, "SSSO_TOKEN=[^%.]+%.[^%.]+%.[^%.]+; Path=/; Expires=1626550390; Secure")
end

function test_post_login_url_with_good_credentials_and_back_url_redirects_to_given_url_with_session_cookie()
  -- given
  ngx.reset_header()
  ngx.reset_var()
  ngx.reset_resp_body()
  ngx.reset_post_var()
  ngx.var.request_method = "POST"
  ngx.var.request_uri = "/ssso/login"
  ngx.post_var.login = "goodlogin"
  ngx.post_var.password = "goodpassword"
  ngx.post_var.back = "/somewhere"
  local r = ng.class__request:current()
  -- when
  local resp = login.answer_request(r)
  -- then
  lu.assertEquals(resp, "302:https://my-domain.tld/somewhere")
  lu.assertNil(ngx.say())
  lu.assertNil(ngx.header["Set-Cookie"].link)
  lu.assertStrMatches(ngx.header["Set-Cookie"].v, "SSSO_TOKEN=[^%.]+%.[^%.]+%.[^%.]+; Path=/; Expires=1626550390; Secure")
end

os.exit(lu.LuaUnit.run())
